#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Network action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from cliff import columns as cliff_columns
from openstack.network.v2 import network as _network
from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import tags as _tag

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common

LOG = logging.getLogger(__name__)


class AdminStateColumn(cliff_columns.FormattableColumn[bool]):
    def human_readable(self) -> str:
        return 'UP' if self._value else 'DOWN'


class RouterExternalColumn(cliff_columns.FormattableColumn[bool]):
    def human_readable(self) -> str:
        return 'External' if self._value else 'Internal'


_formatters = {
    'subnets': format_columns.ListColumn,
    'subnet_ids': format_columns.ListColumn,
    'admin_state_up': AdminStateColumn,
    'is_admin_state_up': AdminStateColumn,
    'router:external': RouterExternalColumn,
    'is_router_external': RouterExternalColumn,
    'availability_zones': format_columns.ListColumn,
    'availability_zone_hints': format_columns.ListColumn,
    'tags': format_columns.ListColumn,
}


def _get_columns_network(
    item: _network.Network,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map = {
        'subnet_ids': 'subnets',
        'is_admin_state_up': 'admin_state_up',
        'is_router_external': 'router:external',
        'is_port_security_enabled': 'port_security_enabled',
        'provider_network_type': 'provider:network_type',
        'provider_physical_network': 'provider:physical_network',
        'provider_segmentation_id': 'provider:segmentation_id',
        'is_shared': 'shared',
        'ipv4_address_scope_id': 'ipv4_address_scope',
        'ipv6_address_scope_id': 'ipv6_address_scope',
        'tags': 'tags',
    }
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


def _get_attrs(
    client_manager: Any, parsed_args: argparse.Namespace
) -> dict[str, Any]:
    attrs = {}
    if parsed_args.name is not None:
        attrs['name'] = parsed_args.name
    if parsed_args.enable:
        attrs['admin_state_up'] = True
    if parsed_args.disable:
        attrs['admin_state_up'] = False
    if parsed_args.share:
        attrs['shared'] = True
    if parsed_args.no_share:
        attrs['shared'] = False
    if parsed_args.enable_port_security:
        attrs['port_security_enabled'] = True
    if parsed_args.disable_port_security:
        attrs['port_security_enabled'] = False

    # "network set" command doesn't support setting project.
    if 'project' in parsed_args and parsed_args.project is not None:
        identity_client = client_manager.sdk_connection.identity
        project_id = identity_common.find_project_id_sdk(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        )
        attrs['project_id'] = project_id

    # "network set" command doesn't support setting availability zone hints.
    if (
        'availability_zone_hints' in parsed_args
        and parsed_args.availability_zone_hints is not None
    ):
        attrs['availability_zone_hints'] = parsed_args.availability_zone_hints

    # set description
    if parsed_args.description is not None:
        attrs['description'] = parsed_args.description

    # set mtu
    if parsed_args.mtu:
        attrs['mtu'] = parsed_args.mtu

    # update_external_network_options
    if parsed_args.internal:
        attrs['router:external'] = False
    if parsed_args.external:
        attrs['router:external'] = True
    if parsed_args.no_default:
        attrs['is_default'] = False
    if parsed_args.default:
        attrs['is_default'] = True
    # Update Provider network options
    if parsed_args.provider_network_type:
        attrs['provider:network_type'] = parsed_args.provider_network_type
    if parsed_args.physical_network:
        attrs['provider:physical_network'] = parsed_args.physical_network
    if parsed_args.segmentation_id:
        attrs['provider:segmentation_id'] = parsed_args.segmentation_id
    if parsed_args.qos_policy is not None:
        network_client = client_manager.network
        _qos_policy = network_client.find_qos_policy(
            parsed_args.qos_policy, ignore_missing=False
        )
        attrs['qos_policy_id'] = _qos_policy.id
    if 'no_qos_policy' in parsed_args and parsed_args.no_qos_policy:
        attrs['qos_policy_id'] = None

    # Set pvlan
    if parsed_args.pvlan:
        attrs['pvlan'] = True
    if parsed_args.no_pvlan:
        attrs['pvlan'] = False

    # Update DNS network options
    if parsed_args.dns_domain is not None:
        attrs['dns_domain'] = parsed_args.dns_domain
    return attrs


def _add_additional_network_options(parser: argparse.ArgumentParser) -> None:
    # Add additional network options

    parser.add_argument(
        '--provider-network-type',
        metavar='<provider-network-type>',
        help=_(
            "The physical mechanism by which the virtual network "
            "is implemented. For example: "
            "flat, geneve, gre, local, vlan or vxlan."
        ),
    )
    parser.add_argument(
        '--provider-physical-network',
        metavar='<provider-physical-network>',
        dest='physical_network',
        help=_(
            "Name of the physical network over which the virtual "
            "network is implemented"
        ),
    )
    parser.add_argument(
        '--provider-segment',
        metavar='<provider-segment>',
        dest='segmentation_id',
        help=_(
            "VLAN ID for VLAN networks or Tunnel ID for "
            "GENEVE/GRE/VXLAN networks"
        ),
    )
    parser.add_argument(
        '--dns-domain',
        metavar='<dns-domain>',
        dest='dns_domain',
        help=_(
            "Set DNS domain for this network "
            "(requires DNS integration extension)"
        ),
    )


class CreateNetwork(command.ShowOne, common.NeutronCommandWithExtraArgs):
    _description = _("Create new network")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name', metavar='<name>', help=_("New network name")
        )
        share_group = parser.add_mutually_exclusive_group()
        share_group.add_argument(
            '--share',
            action='store_true',
            default=None,
            help=_("Share the network between projects"),
        )
        share_group.add_argument(
            '--no-share',
            action='store_true',
            help=_("Do not share the network between projects"),
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=True,
            help=_("Enable network (default)"),
        )
        admin_group.add_argument(
            '--disable',
            action='store_true',
            help=_("Disable network"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("Owner's project (name or ID)"),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Set network description"),
        )
        parser.add_argument(
            '--mtu',
            metavar='<mtu>',
            help=_("Set network mtu"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--availability-zone-hint',
            action='append',
            dest='availability_zone_hints',
            metavar='<availability-zone>',
            help=_(
                "Availability Zone in which to create this network "
                "(Network Availability Zone extension required, "
                "repeat option to set multiple availability zones)"
            ),
        )
        port_security_group = parser.add_mutually_exclusive_group()
        port_security_group.add_argument(
            '--enable-port-security',
            action='store_true',
            help=_(
                "Enable port security by default for ports created on "
                "this network (default)"
            ),
        )
        port_security_group.add_argument(
            '--disable-port-security',
            action='store_true',
            help=_(
                "Disable port security by default for ports created on "
                "this network"
            ),
        )
        external_router_grp = parser.add_mutually_exclusive_group()
        external_router_grp.add_argument(
            '--external',
            action='store_true',
            help=_(
                "The network has an external routing facility that is not "
                "managed by Neutron and can be used. For example: "
                "openstack router set --external-gateway NETWORK "
                "(external-net extension required)"
            ),
        )
        external_router_grp.add_argument(
            '--internal',
            action='store_true',
            help=_("Opposite of '--external' (default)"),
        )
        default_router_grp = parser.add_mutually_exclusive_group()
        default_router_grp.add_argument(
            '--default',
            action='store_true',
            help=_(
                "Specify if this network should be used as the default "
                "external network"
            ),
        )
        default_router_grp.add_argument(
            '--no-default',
            action='store_true',
            help=_(
                "Do not use the network as the default external network "
                "(default)"
            ),
        )
        parser.add_argument(
            '--qos-policy',
            metavar='<qos-policy>',
            help=_("QoS policy to attach to this network (name or ID)"),
        )
        vlan_transparent_grp = parser.add_mutually_exclusive_group()
        vlan_transparent_grp.add_argument(
            '--transparent-vlan',
            action='store_true',
            help=_("Make the network VLAN transparent"),
        )
        vlan_transparent_grp.add_argument(
            '--no-transparent-vlan',
            action='store_true',
            help=_("Do not make the network VLAN transparent"),
        )

        vlan_qinq_grp = parser.add_mutually_exclusive_group()
        vlan_qinq_grp.add_argument(
            '--qinq-vlan',
            action='store_true',
            help=_("Enable VLAN QinQ (S-Tag ethtype 0x8a88) for the network"),
        )
        vlan_qinq_grp.add_argument(
            '--no-qinq-vlan',
            action='store_true',
            help=_("Disable VLAN QinQ (S-Tag ethtype 0x8a88) for the network"),
        )

        pvlan_grp = parser.add_mutually_exclusive_group()
        pvlan_grp.add_argument(
            '--pvlan',
            action='store_true',
            help=_(
                "Enable Private VLAN for the network "
                "(PVLAN extension required)"
            ),
        )
        pvlan_grp.add_argument(
            '--no-pvlan',
            action='store_true',
            help=_(
                "Disable Private VLAN for the network "
                "(PVLAN extension required)"
            ),
        )

        _add_additional_network_options(parser)
        _tag.add_tag_option_to_parser_for_create(parser, _('network'))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        if parsed_args.transparent_vlan:
            attrs['vlan_transparent'] = True
        if parsed_args.no_transparent_vlan:
            attrs['vlan_transparent'] = False

        if parsed_args.qinq_vlan:
            attrs['vlan_qinq'] = True
        if parsed_args.no_qinq_vlan:
            attrs['vlan_qinq'] = False

        if parsed_args.pvlan:
            attrs['pvlan'] = True
        if parsed_args.no_pvlan:
            attrs['pvlan'] = False

        if attrs.get('vlan_transparent') and attrs.get('vlan_qinq'):
            msg = _(
                "--transparent-vlan and --qinq-vlan can not be both enabled "
                "for the network."
            )
            raise exceptions.CommandError(msg)

        if (
            attrs.get('port_security_enabled') is False
            and attrs.get('pvlan') is True
        ):
            msg = _(
                "--disable-port-security and --pvlan can not be used together."
            )
            raise exceptions.CommandError(msg)
        if (
            parsed_args.segmentation_id
            and not parsed_args.provider_network_type
        ):
            msg = _(
                "--provider-segment requires --provider-network-type "
                "to be specified."
            )
            raise exceptions.CommandError(msg)

        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        with common.check_missing_extension_if_error(
            self.app.client_manager.network, attrs
        ):
            obj = client.create_network(**attrs)

        # tags cannot be set when created, so tags need to be set later.
        _tag.update_tags_for_set(client, obj, parsed_args)
        display_columns, columns = _get_columns_network(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)


class DeleteNetwork(command.Command):
    _description = _("Delete network(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar="<network>",
            nargs="+",
            help=_("Network(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for net in parsed_args.network:
            try:
                obj = client.find_network(net, ignore_missing=False)
                client.delete_network(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete network with "
                        "name or ID '%(network)s': %(e)s"
                    ),
                    {'network': net, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.network)
            msg = _("%(result)s of %(total)s networks failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListNetwork(command.Lister):
    _description = _("List networks")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        router_ext_group = parser.add_mutually_exclusive_group()
        router_ext_group.add_argument(
            '--external',
            action='store_true',
            help=_("List only external networks"),
        )
        router_ext_group.add_argument(
            '--internal',
            action='store_true',
            help=_("List only internal networks"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            help=_("List additional fields in output"),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_("List only networks with the specified name"),
        )
        admin_state_group = parser.add_mutually_exclusive_group()
        admin_state_group.add_argument(
            '--enable',
            action='store_true',
            help=_("List only enabled networks"),
        )
        admin_state_group.add_argument(
            '--disable',
            action='store_true',
            help=_("List only disabled networks"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "List only networks with the specified project (name or ID)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        shared_group = parser.add_mutually_exclusive_group()
        shared_group.add_argument(
            '--share',
            action='store_true',
            help=_("List only networks shared between projects"),
        )
        shared_group.add_argument(
            '--no-share',
            action='store_true',
            help=_("List only networks not shared between projects"),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            choices=['ACTIVE', 'BUILD', 'DOWN', 'ERROR'],
            help=_(
                "List only networks with the specified status "
                "('ACTIVE', 'BUILD', 'DOWN', 'ERROR')"
            ),
        )
        parser.add_argument(
            '--provider-network-type',
            metavar='<provider-network-type>',
            choices=['flat', 'geneve', 'gre', 'local', 'vlan', 'vxlan'],
            help=_(
                "List only networks with the specified physical "
                "mechanisms. "
                "The supported options are: flat, geneve, gre, local, "
                "vlan and vxlan."
            ),
        )
        parser.add_argument(
            '--provider-physical-network',
            metavar='<provider-physical-network>',
            dest='physical_network',
            help=_(
                "List only networks with the specified physical network name"
            ),
        )
        parser.add_argument(
            '--provider-segment',
            metavar='<provider-segment>',
            dest='segmentation_id',
            help=_(
                "List only networks with the specified provider segment "
                "ID (VLAN ID for VLAN networks or "
                "Tunnel ID for GENEVE/GRE/VXLAN networks)"
            ),
        )
        parser.add_argument(
            '--agent',
            metavar='<agent-id>',
            dest='agent_id',
            help=_('List only networks hosted the specified agent (ID only)'),
        )
        _tag.add_tag_filtering_option_to_parser(parser, _('networks'))
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        if parsed_args.long:
            columns: tuple[str, ...] = (
                'id',
                'name',
                'status',
                'project_id',
                'is_admin_state_up',
                'is_shared',
                'subnet_ids',
                'provider_network_type',
                'is_router_external',
                'availability_zones',
                'tags',
            )
            column_headers: tuple[str, ...] = (
                'ID',
                'Name',
                'Status',
                'Project',
                'State',
                'Shared',
                'Subnets',
                'Network Type',
                'Router Type',
                'Availability Zones',
                'Tags',
            )
        else:
            columns = ('id', 'name', 'subnet_ids')
            column_headers = (
                'ID',
                'Name',
                'Subnets',
            )

        if parsed_args.agent_id:
            columns = ('id', 'name', 'subnet_ids')
            column_headers = (
                'ID',
                'Name',
                'Subnets',
            )
            client = self.app.client_manager.network
            dhcp_agent = client.get_agent(parsed_args.agent_id)

            return (
                column_headers,
                (
                    utils.get_item_properties(
                        s,
                        columns,
                        formatters=_formatters,
                    )
                    for s in client.dhcp_agent_hosting_networks(dhcp_agent)
                ),
            )

        args: dict[str, object] = {}

        if parsed_args.external:
            args['router:external'] = True
            args['is_router_external'] = True
        elif parsed_args.internal:
            args['router:external'] = False
            args['is_router_external'] = False

        if parsed_args.name is not None:
            args['name'] = parsed_args.name

        if parsed_args.enable:
            args['admin_state_up'] = True
            args['is_admin_state_up'] = True
        elif parsed_args.disable:
            args['admin_state_up'] = False
            args['is_admin_state_up'] = False

        if parsed_args.project:
            identity_client = self.app.client_manager.sdk_connection.identity
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )
            args['project_id'] = project_id

        if parsed_args.share:
            args['shared'] = True
            args['is_shared'] = True
        elif parsed_args.no_share:
            args['shared'] = False
            args['is_shared'] = False

        if parsed_args.status:
            args['status'] = parsed_args.status

        if parsed_args.provider_network_type:
            args['provider:network_type'] = parsed_args.provider_network_type
            args['provider_network_type'] = parsed_args.provider_network_type
        if parsed_args.physical_network:
            args['provider:physical_network'] = parsed_args.physical_network
            args['provider_physical_network'] = parsed_args.physical_network
        if parsed_args.segmentation_id:
            args['provider:segmentation_id'] = parsed_args.segmentation_id
            args['provider_segmentation_id'] = parsed_args.segmentation_id

        if parsed_args.marker is not None:
            args['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            args['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            args['max_items'] = parsed_args.max_items

        _tag.get_tag_filtering_args(parsed_args, args)

        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters=_formatters,
                )
                for s in client.networks(**args)
            ),
        )


class SetNetwork(common.NeutronCommandWithExtraArgs):
    _description = _("Set network properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar="<network>",
            help=_("Network to modify (name or ID)"),
        )
        parser.add_argument(
            '--name', metavar='<name>', help=_("Set network name")
        )
        admin_group = parser.add_mutually_exclusive_group()
        admin_group.add_argument(
            '--enable',
            action='store_true',
            default=None,
            help=_("Enable network"),
        )
        admin_group.add_argument(
            '--disable', action='store_true', help=_("Disable network")
        )
        share_group = parser.add_mutually_exclusive_group()
        share_group.add_argument(
            '--share',
            action='store_true',
            default=None,
            help=_("Share the network between projects"),
        )
        share_group.add_argument(
            '--no-share',
            action='store_true',
            help=_("Do not share the network between projects"),
        )
        parser.add_argument(
            '--description',
            metavar="<description>",
            help=_("Set network description"),
        )
        parser.add_argument(
            '--mtu', metavar="<mtu>", help=_("Set network mtu")
        )
        port_security_group = parser.add_mutually_exclusive_group()
        port_security_group.add_argument(
            '--enable-port-security',
            action='store_true',
            help=_(
                "Enable port security by default for ports created on "
                "this network"
            ),
        )
        port_security_group.add_argument(
            '--disable-port-security',
            action='store_true',
            help=_(
                "Disable port security by default for ports created on "
                "this network"
            ),
        )
        external_router_grp = parser.add_mutually_exclusive_group()
        external_router_grp.add_argument(
            '--external',
            action='store_true',
            help=_(
                "The network has an external routing facility that is not "
                "managed by Neutron and can be used. For example: "
                "openstack router set --external-gateway NETWORK "
                "(external-net extension required)."
            ),
        )
        external_router_grp.add_argument(
            '--internal',
            action='store_true',
            help=_("Opposite of '--external'"),
        )
        default_router_grp = parser.add_mutually_exclusive_group()
        default_router_grp.add_argument(
            '--default',
            action='store_true',
            help=_("Set the network as the default external network"),
        )
        default_router_grp.add_argument(
            '--no-default',
            action='store_true',
            help=_("Do not use the network as the default external network"),
        )
        qos_group = parser.add_mutually_exclusive_group()
        qos_group.add_argument(
            '--qos-policy',
            metavar='<qos-policy>',
            help=_("QoS policy to attach to this network (name or ID)"),
        )
        qos_group.add_argument(
            '--no-qos-policy',
            action='store_true',
            help=_("Remove the QoS policy attached to this network"),
        )
        pvlan_grp = parser.add_mutually_exclusive_group()
        pvlan_grp.add_argument(
            '--pvlan',
            action='store_true',
            help=_(
                "Enable Private VLAN for the network. PVLAN extension "
                "required."
            ),
        )
        pvlan_grp.add_argument(
            '--no-pvlan',
            action='store_true',
            help=_(
                "Disable Private VLAN for the network (Default). "
                "PVLAN extension required."
            ),
        )
        _tag.add_tag_option_to_parser_for_set(parser, _('network'))
        _add_additional_network_options(parser)
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_network(parsed_args.network, ignore_missing=False)

        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        if (
            attrs.get('port_security_enabled') is False
            and attrs.get('pvlan') is True
        ):
            msg = _(
                "--disable-port-security and --pvlan can not be used together."
            )
            raise exceptions.CommandError(msg)
        if attrs:
            with common.check_missing_extension_if_error(
                self.app.client_manager.network, attrs
            ):
                client.update_network(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_set(client, obj, parsed_args)


class ShowNetwork(command.ShowOne):
    _description = _("Show network details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar="<network>",
            help=_("Network to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_network(parsed_args.network, ignore_missing=False)
        display_columns, columns = _get_columns_network(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)


class UnsetNetwork(common.NeutronUnsetCommandWithExtraArgs):
    _description = _("Unset network properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'network',
            metavar="<network>",
            help=_("Network to modify (name or ID)"),
        )
        _tag.add_tag_option_to_parser_for_unset(parser, _('network'))
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_network(parsed_args.network, ignore_missing=False)

        attrs = self._parse_extra_properties(parsed_args.extra_properties)
        if attrs:
            client.update_network(obj, **attrs)

        # tags is a subresource and it needs to be updated separately.
        _tag.update_tags_for_unset(client, obj, parsed_args)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume V3 Messages implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack.block_storage.v3 import message as _message
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

LOG = logging.getLogger(__name__)


def _format_message(
    message: _message.Message,
) -> tuple[tuple[str, ...], tuple[Any, ...]]:
    column_headers = (
        'Created At',
        'Event ID',
        'Guaranteed Until',
        'ID',
        'Message Level',
        'Request ID',
        'Resource Type',
        'Resource UUID',
        'User Message',
    )
    columns = (
        'created_at',
        'event_id',
        'guaranteed_until',
        'id',
        'message_level',
        'request_id',
        'resource_type',
        'resource_uuid',
        'user_message',
    )
    return column_headers, utils.get_item_properties(message, columns)


class DeleteMessage(command.Command):
    _description = _('Delete a volume failure message')

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'message_ids',
            metavar='<message-id>',
            nargs='+',
            help=_('Message(s) to delete (ID)'),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.3'):
            msg = _(
                "--os-volume-api-version 3.3 or greater is required to "
                "support the 'volume message delete' command"
            )
            raise exceptions.CommandError(msg)

        errors = 0
        for message_id in parsed_args.message_ids:
            try:
                volume_client.delete_message(message_id, ignore_missing=False)
            except Exception:
                LOG.error(_('Failed to delete message: %s'), message_id)
                errors += 1

        if errors > 0:
            total = len(parsed_args.message_ids)
            msg = _('Failed to delete %(errors)s of %(total)s messages.') % {
                'errors': errors,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListMessages(command.Lister):
    _description = _('List volume failure messages')

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Filter results by project (name or ID) (admin only)'),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        pagination.add_marker_pagination_option_to_parser(
            parser, include_max_items=False
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.3'):
            msg = _(
                "--os-volume-api-version 3.3 or greater is required to "
                "support the 'volume message list' command"
            )
            raise exceptions.CommandError(msg)

        column_headers = (
            'ID',
            'Event ID',
            'Resource Type',
            'Resource UUID',
            'Message Level',
            'User Message',
            'Request ID',
            'Created At',
            'Guaranteed Until',
        )
        columns = (
            'id',
            'event_id',
            'resource_type',
            'resource_uuid',
            'message_level',
            'user_message',
            'request_id',
            'created_at',
            'guaranteed_until',
        )

        project_id = None
        if parsed_args.project:
            project_id = identity_common.find_project_id_sdk(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            )

        data = volume_client.messages(
            project_id=project_id,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
        )

        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in data),
        )


class ShowMessage(command.ShowOne):
    _description = _('Show a volume failure message')

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'message_id',
            metavar='<message-id>',
            help=_('Message to show (ID).'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )

        if not sdk_utils.supports_microversion(volume_client, '3.3'):
            msg = _(
                "--os-volume-api-version 3.3 or greater is required to "
                "support the 'volume message show' command"
            )
            raise exceptions.CommandError(msg)

        message = volume_client.get_message(parsed_args.message_id)

        return _format_message(message)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import logging
from typing import Any

from osc_lib import utils

from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

# global variables used when building the shell
DEFAULT_API_VERSION = '2.1'
API_VERSION_OPTION = 'os_compute_api_version'
API_NAME = 'compute'
API_VERSIONS = ('2', '2.1')


def make_client(instance: Any) -> Any:
    """Returns a compute service client."""
    LOG.debug(
        'Compute client initialized using OpenStack SDK: %s',
        instance.sdk_connection.compute,
    )
    return instance.sdk_connection.compute


def build_option_parser(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Hook to add global options"""
    parser.add_argument(
        '--os-compute-api-version',
        metavar='<compute-api-version>',
        default=utils.env('OS_COMPUTE_API_VERSION'),
        help=_("Compute API version, default=%s (Env: OS_COMPUTE_API_VERSION)")
        % DEFAULT_API_VERSION,
    )
    return parser


def check_api_version(check_version: str) -> bool:
    # SDK supports auto-negotiation for us: always return True
    return True



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Volume v3 Backup action implementations"""

import argparse
import logging
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import utils as sdk_utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)


class ExportBackupRecord(command.ShowOne):
    _description = _(
        """Export volume backup details.

Backup information can be imported into a new service instance to be able to
restore."""
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup",
            metavar="<backup>",
            help=_("Backup to export (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        backup = volume_client.find_backup(
            parsed_args.backup, ignore_missing=False
        )
        backup_data = volume_client.export_backup(backup)['backup-record']

        # We only want to show "friendly" display names, but also want to keep
        # json structure compatibility with cinderclient
        if parsed_args.formatter == 'table':
            backup_data['Backup Service'] = backup_data.pop('backup_service')
            backup_data['Metadata'] = backup_data.pop('backup_url')

        col_headers, col_data = zip(*sorted(backup_data.items()))
        return col_headers, col_data


class ImportBackupRecord(command.ShowOne):
    _description = _(
        """Import volume backup details.

Exported backup details contain the metadata necessary to restore to a new or
rebuilt service instance"""
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "backup_service",
            metavar="<backup_service>",
            help=_("Backup service containing the backup."),
        )
        parser.add_argument(
            "backup_metadata",
            metavar="<backup_metadata>",
            help=_("Encoded backup metadata from export."),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        volume_client = sdk_utils.ensure_service_version(
            self.app.client_manager.volume, '3'
        )
        backup_data = volume_client.import_backup(
            parsed_args.backup_service, parsed_args.backup_metadata
        )
        backup_data.pop('links', None)
        col_headers, col_data = zip(*sorted(backup_data.items()))
        return col_headers, col_data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest.mock import call

from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import project as _project
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network import utils as network_utils
from openstackclient.network.v2 import security_group_rule
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestSecurityGroupRuleNetwork(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()


class TestCreateSecurityGroupRuleNetwork(TestSecurityGroupRuleNetwork):
    project = sdk_fakes.generate_fake_resource(_project.Project)
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    # The security group rule to be created.
    _security_group_rule = None

    # The security group that will contain the rule created.
    _security_group = network_fakes.create_one_security_group()

    # The address group to be used in security group rules
    _address_group = network_fakes.create_one_address_group()

    expected_columns = (
        'created_at',
        'description',
        'direction',
        'ether_type',
        'id',
        'port_range_max',
        'port_range_min',
        'project_id',
        'protocol',
        'remote_address_group_id',
        'remote_group_id',
        'remote_ip_prefix',
        'revision_number',
        'security_group_id',
        'updated_at',
    )

    expected_data = None

    def _setup_security_group_rule(self, attrs=None):
        self._security_group_rule = (
            network_fakes.create_one_security_group_rule(attrs)
        )
        self.network_client.create_security_group_rule.return_value = (
            self._security_group_rule
        )

        self.expected_data = (
            self._security_group_rule.created_at,
            self._security_group_rule.description,
            self._security_group_rule.direction,
            self._security_group_rule.ether_type,
            self._security_group_rule.id,
            self._security_group_rule.port_range_max,
            self._security_group_rule.port_range_min,
            self._security_group_rule.project_id,
            self._security_group_rule.protocol,
            self._security_group_rule.remote_address_group_id,
            self._security_group_rule.remote_group_id,
            self._security_group_rule.remote_ip_prefix,
            self._security_group_rule.revision_number,
            self._security_group_rule.security_group_id,
            self._security_group_rule.updated_at,
        )

    def setUp(self):
        super().setUp()

        self.network_client.find_security_group.return_value = (
            self._security_group
        )

        self.network_client.find_address_group.return_value = (
            self._address_group
        )

        self.identity_sdk_client.find_project.return_value = self.project

        # Get the command object to test
        self.cmd = security_group_rule.CreateSecurityGroupRule(self.app, None)

    def test_create_no_options(self):
        self.assertRaises(
            tests_utils.ParserException, self.check_parser, self.cmd, [], []
        )

    def test_create_all_remote_options(self):
        arglist = [
            '--remote-ip',
            '10.10.0.0/24',
            '--remote-group',
            self._security_group.id,
            '--remote-address-group',
            self._address_group.id,
            self._security_group.id,
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            [],
        )

    def test_create_bad_ethertype(self):
        arglist = [
            '--ethertype',
            'foo',
            self._security_group.id,
        ]
        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            [],
        )

    def test_lowercase_ethertype(self):
        arglist = [
            '--ethertype',
            'ipv4',
            self._security_group.id,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.assertEqual('IPv4', parsed_args.ethertype)

    def test_lowercase_v6_ethertype(self):
        arglist = [
            '--ethertype',
            'ipv6',
            self._security_group.id,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.assertEqual('IPv6', parsed_args.ethertype)

    def test_proper_case_ethertype(self):
        arglist = [
            '--ethertype',
            'IPv6',
            self._security_group.id,
        ]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.assertEqual('IPv6', parsed_args.ethertype)

    def test_create_all_port_range_options(self):
        arglist = [
            '--dst-port',
            '80:80',
            '--icmp-type',
            '3',
            '--icmp-code',
            '1',
            self._security_group.id,
        ]
        verifylist = [
            ('dst_port', (80, 80)),
            ('icmp_type', 3),
            ('icmp_code', 1),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_default_rule(self):
        self._setup_security_group_rule(
            {
                'protocol': 'tcp',
                'port_range_max': 443,
                'port_range_min': 443,
            }
        )
        arglist = [
            '--protocol',
            'tcp',
            '--dst-port',
            str(self._security_group_rule.port_range_min),
            self._security_group.id,
        ]
        verifylist = [
            (
                'dst_port',
                (
                    self._security_group_rule.port_range_min,
                    self._security_group_rule.port_range_max,
                ),
            ),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_max': self._security_group_rule.port_range_max,
                'port_range_min': self._security_group_rule.port_range_min,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_protocol_option(self):
        self._setup_security_group_rule(
            {
                'protocol': 'icmp',
                'remote_ip_prefix': '10.0.2.0/24',
            }
        )
        arglist = [
            '--protocol',
            self._security_group_rule.protocol,
            '--remote-ip',
            self._security_group_rule.remote_ip_prefix,
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', self._security_group_rule.protocol),
            ('remote_ip', self._security_group_rule.remote_ip_prefix),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_protocol_any(self):
        self._setup_security_group_rule(
            {
                'protocol': None,
                'remote_ip_prefix': '10.0.2.0/24',
            }
        )
        arglist = [
            '--protocol',
            'any',
            '--remote-ip',
            self._security_group_rule.remote_ip_prefix,
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', 'any'),
            ('remote_ip', self._security_group_rule.remote_ip_prefix),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_remote_address_group(self):
        self._setup_security_group_rule(
            {
                'protocol': 'icmp',
                'remote_address_group_id': self._address_group.id,
            }
        )
        arglist = [
            '--protocol',
            'icmp',
            '--remote-address-group',
            self._address_group.name,
            self._security_group.id,
        ]
        verifylist = [
            ('remote_address_group', self._address_group.name),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_address_group_id': self._address_group.id,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_remote_group(self):
        self._setup_security_group_rule(
            {
                'protocol': 'tcp',
                'port_range_max': 22,
                'port_range_min': 22,
            }
        )
        arglist = [
            '--protocol',
            'tcp',
            '--dst-port',
            str(self._security_group_rule.port_range_min),
            '--ingress',
            '--remote-group',
            self._security_group.name,
            self._security_group.id,
        ]
        verifylist = [
            (
                'dst_port',
                (
                    self._security_group_rule.port_range_min,
                    self._security_group_rule.port_range_max,
                ),
            ),
            ('ingress', True),
            ('remote_group', self._security_group.name),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_max': self._security_group_rule.port_range_max,
                'port_range_min': self._security_group_rule.port_range_min,
                'protocol': self._security_group_rule.protocol,
                'remote_group_id': self._security_group.id,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_source_group(self):
        self._setup_security_group_rule(
            {
                'remote_group_id': self._security_group.id,
            }
        )
        arglist = [
            '--ingress',
            '--remote-group',
            self._security_group.name,
            self._security_group.id,
        ]
        verifylist = [
            ('ingress', True),
            ('remote_group', self._security_group.name),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_group_id': self._security_group_rule.remote_group_id,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_source_ip(self):
        self._setup_security_group_rule(
            {
                'protocol': 'icmp',
                'remote_ip_prefix': '10.0.2.0/24',
            }
        )
        arglist = [
            '--protocol',
            self._security_group_rule.protocol,
            '--remote-ip',
            self._security_group_rule.remote_ip_prefix,
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', self._security_group_rule.protocol),
            ('remote_ip', self._security_group_rule.remote_ip_prefix),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_remote_ip(self):
        self._setup_security_group_rule(
            {
                'protocol': 'icmp',
                'remote_ip_prefix': '10.0.2.0/24',
            }
        )
        arglist = [
            '--protocol',
            self._security_group_rule.protocol,
            '--remote-ip',
            self._security_group_rule.remote_ip_prefix,
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', self._security_group_rule.protocol),
            ('remote_ip', self._security_group_rule.remote_ip_prefix),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_network_options(self):
        self._setup_security_group_rule(
            {
                'direction': 'egress',
                'ether_type': 'IPv6',
                'port_range_max': 443,
                'port_range_min': 443,
                'protocol': '6',
                'remote_group_id': None,
                'remote_ip_prefix': '::/0',
            }
        )
        arglist = [
            '--dst-port',
            str(self._security_group_rule.port_range_min),
            '--egress',
            '--ethertype',
            self._security_group_rule.ether_type,
            '--project',
            self.project.name,
            '--project-domain',
            self.domain.name,
            '--protocol',
            self._security_group_rule.protocol,
            self._security_group.id,
        ]
        verifylist = [
            (
                'dst_port',
                (
                    self._security_group_rule.port_range_min,
                    self._security_group_rule.port_range_max,
                ),
            ),
            ('egress', True),
            ('ethertype', self._security_group_rule.ether_type),
            ('project', self.project.name),
            ('project_domain', self.domain.name),
            ('protocol', self._security_group_rule.protocol),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_max': self._security_group_rule.port_range_max,
                'port_range_min': self._security_group_rule.port_range_min,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
                'project_id': self.project.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_tcp_with_icmp_type(self):
        arglist = [
            '--protocol',
            'tcp',
            '--icmp-type',
            '15',
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', 'tcp'),
            ('icmp_type', 15),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_icmp_code(self):
        arglist = [
            '--protocol',
            '1',
            '--icmp-code',
            '1',
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', '1'),
            ('icmp_code', 1),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_icmp_code_zero(self):
        self._setup_security_group_rule(
            {
                'port_range_min': 15,
                'port_range_max': 0,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--protocol',
            self._security_group_rule.protocol,
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--icmp-code',
            str(self._security_group_rule.port_range_max),
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', self._security_group_rule.protocol),
            ('icmp_code', self._security_group_rule.port_range_max),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_code_greater_than_zero(self):
        self._setup_security_group_rule(
            {
                'port_range_min': 15,
                'port_range_max': 18,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--protocol',
            self._security_group_rule.protocol,
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--icmp-code',
            str(self._security_group_rule.port_range_max),
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', self._security_group_rule.protocol),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('icmp_code', self._security_group_rule.port_range_max),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_code_negative_value(self):
        self._setup_security_group_rule(
            {
                'port_range_min': 15,
                'port_range_max': None,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--protocol',
            self._security_group_rule.protocol,
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--icmp-code',
            '-2',
            self._security_group.id,
        ]
        verifylist = [
            ('protocol', self._security_group_rule.protocol),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('icmp_code', -2),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type(self):
        self._setup_security_group_rule(
            {
                'port_range_min': 15,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--protocol',
            self._security_group_rule.protocol,
            self._security_group.id,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._security_group_rule.protocol),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_min': self._security_group_rule.port_range_min,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type_zero(self):
        self._setup_security_group_rule(
            {
                'port_range_min': 0,
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--protocol',
            self._security_group_rule.protocol,
            self._security_group.id,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._security_group_rule.protocol),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_min': self._security_group_rule.port_range_min,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type_greater_than_zero(self):
        self._setup_security_group_rule(
            {
                'port_range_min': 13,  # timestamp
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--protocol',
            self._security_group_rule.protocol,
            self._security_group.id,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._security_group_rule.protocol),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_min': self._security_group_rule.port_range_min,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmp_type_negative_value(self):
        self._setup_security_group_rule(
            {
                'port_range_min': None,  # timestamp
                'protocol': 'icmp',
                'remote_ip_prefix': '0.0.0.0/0',
            }
        )
        arglist = [
            '--icmp-type',
            '-13',
            '--protocol',
            self._security_group_rule.protocol,
            self._security_group.id,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', -13),
            ('icmp_code', None),
            ('protocol', self._security_group_rule.protocol),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_ipv6_icmp_type_code(self):
        self._setup_security_group_rule(
            {
                'ether_type': 'IPv6',
                'port_range_min': 139,
                'port_range_max': 2,
                'protocol': 'ipv6-icmp',
                'remote_ip_prefix': '::/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--icmp-code',
            str(self._security_group_rule.port_range_max),
            '--protocol',
            self._security_group_rule.protocol,
            self._security_group.id,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('icmp_code', self._security_group_rule.port_range_max),
            ('protocol', self._security_group_rule.protocol),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_min': self._security_group_rule.port_range_min,
                'port_range_max': self._security_group_rule.port_range_max,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_icmpv6_type(self):
        self._setup_security_group_rule(
            {
                'ether_type': 'IPv6',
                'port_range_min': 139,
                'protocol': 'icmpv6',
                'remote_ip_prefix': '::/0',
            }
        )
        arglist = [
            '--icmp-type',
            str(self._security_group_rule.port_range_min),
            '--protocol',
            self._security_group_rule.protocol,
            self._security_group.id,
        ]
        verifylist = [
            ('dst_port', None),
            ('icmp_type', self._security_group_rule.port_range_min),
            ('icmp_code', None),
            ('protocol', self._security_group_rule.protocol),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'port_range_min': self._security_group_rule.port_range_min,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)

    def test_create_with_description(self):
        self._setup_security_group_rule(
            {
                'description': 'Setting SGR',
            }
        )
        arglist = [
            '--description',
            self._security_group_rule.description,
            self._security_group.id,
        ]
        verifylist = [
            ('description', self._security_group_rule.description),
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_security_group_rule.assert_called_once_with(
            **{
                'description': self._security_group_rule.description,
                'direction': self._security_group_rule.direction,
                'ethertype': self._security_group_rule.ether_type,
                'protocol': self._security_group_rule.protocol,
                'remote_ip_prefix': self._security_group_rule.remote_ip_prefix,
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)


class TestDeleteSecurityGroupRuleNetwork(TestSecurityGroupRuleNetwork):
    # The security group rules to be deleted.
    _security_group_rules = network_fakes.create_security_group_rules(count=2)

    def setUp(self):
        super().setUp()

        self.network_client.delete_security_group_rule.return_value = None

        self.network_client.find_security_group_rule = (
            network_fakes.get_security_group_rules(self._security_group_rules)
        )

        # Get the command object to test
        self.cmd = security_group_rule.DeleteSecurityGroupRule(self.app, None)

    def test_security_group_rule_delete(self):
        arglist = [
            self._security_group_rules[0].id,
        ]
        verifylist = [
            ('rule', [self._security_group_rules[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_security_group_rule.assert_called_once_with(
            self._security_group_rules[0]
        )
        self.assertIsNone(result)

    def test_multi_security_group_rules_delete(self):
        arglist = []
        verifylist = []

        for s in self._security_group_rules:
            arglist.append(s.id)
        verifylist = [
            ('rule', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for s in self._security_group_rules:
            calls.append(call(s))
        self.network_client.delete_security_group_rule.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_security_group_rules_delete_with_exception(self):
        arglist = [
            self._security_group_rules[0].id,
            'unexist_rule',
        ]
        verifylist = [
            ('rule', [self._security_group_rules[0].id, 'unexist_rule']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [
            self._security_group_rules[0],
            exceptions.CommandError,
        ]
        self.network_client.find_security_group_rule.side_effect = (
            find_mock_result
        )

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 rules failed to delete.', str(e))

        self.network_client.find_security_group_rule.assert_any_call(
            self._security_group_rules[0].id, ignore_missing=False
        )
        self.network_client.find_security_group_rule.assert_any_call(
            'unexist_rule', ignore_missing=False
        )
        self.network_client.delete_security_group_rule.assert_called_once_with(
            self._security_group_rules[0]
        )


class TestListSecurityGroupRuleNetwork(TestSecurityGroupRuleNetwork):
    # The security group to hold the rules.
    _security_group = network_fakes.create_one_security_group()

    # The security group rule to be listed.
    _security_group_rule_tcp = network_fakes.create_one_security_group_rule(
        {
            'protocol': 'tcp',
            'port_range_max': 80,
            'port_range_min': 80,
            'security_group_id': _security_group.id,
        }
    )
    _security_group_rule_icmp = network_fakes.create_one_security_group_rule(
        {
            'protocol': 'icmp',
            'remote_ip_prefix': '10.0.2.0/24',
            'security_group_id': _security_group.id,
        }
    )
    _security_group.security_group_rules = [
        dict(_security_group_rule_tcp),
        dict(_security_group_rule_icmp),
    ]
    _security_group_rules = [
        _security_group_rule_tcp,
        _security_group_rule_icmp,
    ]

    expected_columns_with_group = (
        'ID',
        'IP Protocol',
        'Ethertype',
        'IP Range',
        'Port Range',
        'Direction',
        'Remote Security Group',
        'Remote Address Group',
    )
    expected_columns_no_group = (
        'ID',
        'IP Protocol',
        'Ethertype',
        'IP Range',
        'Port Range',
        'Direction',
        'Remote Security Group',
        'Remote Address Group',
        'Security Group',
    )

    expected_data_with_group = []
    expected_data_no_group = []
    for _security_group_rule in _security_group_rules:
        expected_data_with_group.append(
            (
                _security_group_rule.id,
                _security_group_rule.protocol,
                _security_group_rule.ether_type,
                _security_group_rule.remote_ip_prefix,
                network_utils.format_network_port_range(_security_group_rule),
                _security_group_rule.direction,
                _security_group_rule.remote_group_id,
                _security_group_rule.remote_address_group_id,
            )
        )
        expected_data_no_group.append(
            (
                _security_group_rule.id,
                _security_group_rule.protocol,
                _security_group_rule.ether_type,
                _security_group_rule.remote_ip_prefix,
                network_utils.format_network_port_range(_security_group_rule),
                _security_group_rule.direction,
                _security_group_rule.remote_group_id,
                _security_group_rule.remote_address_group_id,
                _security_group_rule.security_group_id,
            )
        )

    def setUp(self):
        super().setUp()

        self.network_client.find_security_group.return_value = (
            self._security_group
        )

        self.network_client.security_group_rules.return_value = (
            self._security_group_rules
        )

        # Get the command object to test
        self.cmd = security_group_rule.ListSecurityGroupRule(self.app, None)

    def test_list_default(self):
        self._security_group_rule_tcp.port_range_min = 80
        parsed_args = self.check_parser(self.cmd, [], [])

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.security_group_rules.assert_called_once_with(**{})
        self.assertEqual(self.expected_columns_no_group, columns)
        self.assertEqual(self.expected_data_no_group, list(data))

    def test_list_with_pagination(self):
        self._security_group_rule_tcp.port_range_min = 80
        arglist = [
            '--marker',
            self._security_group_rules[0].id,
            '--limit',
            '1',
        ]
        verifylist = [
            ('marker', self._security_group_rules[0].id),
            ('limit', 1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.security_group_rules.assert_called_once_with(
            **{
                'marker': self._security_group_rules[0].id,
                'limit': 1,
            }
        )
        self.assertEqual(self.expected_columns_no_group, columns)
        self.assertEqual(self.expected_data_no_group, list(data))

    def test_list_with_group(self):
        self._security_group_rule_tcp.port_range_min = 80
        arglist = [
            self._security_group.id,
        ]
        verifylist = [
            ('group', self._security_group.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.security_group_rules.assert_called_once_with(
            **{
                'security_group_id': self._security_group.id,
            }
        )
        self.assertEqual(self.expected_columns_with_group, columns)
        self.assertEqual(self.expected_data_with_group, list(data))

    def test_list_with_protocol(self):
        self._security_group_rule_tcp.port_range_min = 80
        arglist = [
            '--protocol',
            'tcp',
        ]
        verifylist = [
            ('protocol', 'tcp'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.security_group_rules.assert_called_once_with(
            **{
                'protocol': 'tcp',
            }
        )
        self.assertEqual(self.expected_columns_no_group, columns)
        self.assertEqual(self.expected_data_no_group, list(data))

    def test_list_with_ingress(self):
        self._security_group_rule_tcp.port_range_min = 80
        arglist = [
            '--ingress',
        ]
        verifylist = [
            ('ingress', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.security_group_rules.assert_called_once_with(
            **{
                'direction': 'ingress',
            }
        )
        self.assertEqual(self.expected_columns_no_group, columns)
        self.assertEqual(self.expected_data_no_group, list(data))

    def test_list_with_wrong_egress(self):
        self._security_group_rule_tcp.port_range_min = 80
        arglist = [
            '--egress',
        ]
        verifylist = [
            ('egress', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.security_group_rules.assert_called_once_with(
            **{
                'direction': 'egress',
            }
        )
        self.assertEqual(self.expected_columns_no_group, columns)
        self.assertEqual(self.expected_data_no_group, list(data))

    def test_list_with_project(self):
        project = sdk_fakes.generate_fake_resource(_project.Project)
        self._security_group_rule_tcp.port_range_min = 80
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            '--project',
            project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.security_group_rules.assert_called_once_with(
            **filters
        )
        self.assertEqual(self.expected_columns_no_group, columns)
        self.assertEqual(self.expected_data_no_group, list(data))

    def test_list_with_project_domain(self):
        project = sdk_fakes.generate_fake_resource(_project.Project)
        self._security_group_rule_tcp.port_range_min = 80
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            '--project',
            project.id,
            '--project-domain',
            project.domain_id,
        ]
        verifylist = [
            ('project', project.id),
            ('project_domain', project.domain_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.security_group_rules.assert_called_once_with(
            **filters
        )
        self.assertEqual(self.expected_columns_no_group, columns)
        self.assertEqual(self.expected_data_no_group, list(data))


class TestShowSecurityGroupRuleNetwork(TestSecurityGroupRuleNetwork):
    # The security group rule to be shown.
    _security_group_rule = network_fakes.create_one_security_group_rule()

    columns = (
        'created_at',
        'description',
        'direction',
        'ether_type',
        'id',
        'port_range_max',
        'port_range_min',
        'project_id',
        'protocol',
        'remote_address_group_id',
        'remote_group_id',
        'remote_ip_prefix',
        'revision_number',
        'security_group_id',
        'updated_at',
    )

    data = (
        _security_group_rule.created_at,
        _security_group_rule.description,
        _security_group_rule.direction,
        _security_group_rule.ether_type,
        _security_group_rule.id,
        _security_group_rule.port_range_max,
        _security_group_rule.port_range_min,
        _security_group_rule.project_id,
        _security_group_rule.protocol,
        _security_group_rule.remote_address_group_id,
        _security_group_rule.remote_group_id,
        _security_group_rule.remote_ip_prefix,
        _security_group_rule.revision_number,
        _security_group_rule.security_group_id,
        _security_group_rule.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.network_client.find_security_group_rule.return_value = (
            self._security_group_rule
        )

        # Get the command object to test
        self.cmd = security_group_rule.ShowSecurityGroupRule(self.app, None)

    def test_show_no_options(self):
        self.assertRaises(
            tests_utils.ParserException, self.check_parser, self.cmd, [], []
        )

    def test_show_all_options(self):
        arglist = [
            self._security_group_rule.id,
        ]
        verifylist = [
            ('rule', self._security_group_rule.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_security_group_rule.assert_called_once_with(
            self._security_group_rule.id, ignore_missing=False
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from cliff import columns as cliff_columns
from openstack.network.v2 import qos_policy as _qos_policy
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common
from openstackclient.network import common

LOG = logging.getLogger(__name__)


class RulesColumn(cliff_columns.FormattableColumn[Any]):
    def human_readable(self) -> str:
        return '\n'.join(str(v) for v in self._value)


_formatters = {
    'rules': RulesColumn,
}


def _get_columns(
    item: _qos_policy.QoSPolicy,
) -> tuple[tuple[str, ...], tuple[str, ...]]:
    column_map = {
        'is_shared': 'shared',
    }
    hidden_columns = ['location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


def _get_attrs(
    client_manager: Any, parsed_args: argparse.Namespace
) -> dict[str, Any]:
    attrs = {}
    if 'name' in parsed_args and parsed_args.name is not None:
        attrs['name'] = parsed_args.name
    if 'description' in parsed_args and parsed_args.description is not None:
        attrs['description'] = parsed_args.description
    if parsed_args.share:
        attrs['shared'] = True
    if parsed_args.no_share:
        attrs['shared'] = False
    # NOTE(ralonsoh): 'default' and 'no_default' parameters are defined only in
    #                create and set commands context only.
    if 'default' in parsed_args and parsed_args.default:
        attrs['is_default'] = True
    if 'no_default' in parsed_args and parsed_args.no_default:
        attrs['is_default'] = False
    # NOTE(ralonsoh): 'project' parameter is defined only in create and list
    #                commands context only.
    if 'project' in parsed_args and parsed_args.project is not None:
        identity_client = client_manager.sdk_connection.identity
        project_id = identity_common.find_project_id_sdk(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        )
        attrs['project_id'] = project_id

    return attrs


class CreateNetworkQosPolicy(
    command.ShowOne, common.NeutronCommandWithExtraArgs
):
    _description = _("Create a QoS policy")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name', metavar='<name>', help=_("Name of QoS policy to create")
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Description of the QoS policy"),
        )
        share_group = parser.add_mutually_exclusive_group()
        share_group.add_argument(
            '--share',
            action='store_true',
            default=None,
            help=_("Make the QoS policy accessible by other projects"),
        )
        share_group.add_argument(
            '--no-share',
            action='store_true',
            help=_(
                "Make the QoS policy not accessible by other projects "
                "(default)"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_("Owner's project (name or ID)"),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        default_group = parser.add_mutually_exclusive_group()
        default_group.add_argument(
            '--default',
            action='store_true',
            help=_("Set this as a default network QoS policy"),
        )
        default_group.add_argument(
            '--no-default',
            action='store_true',
            help=_("Set this as a non-default network QoS policy"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        obj = client.create_qos_policy(**attrs)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})
        return (display_columns, data)


class DeleteNetworkQosPolicy(command.Command):
    _description = _("Delete Qos Policy(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar="<qos-policy>",
            nargs="+",
            help=_("QoS policy(s) to delete (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        result = 0

        for policy in parsed_args.policy:
            try:
                obj = client.find_qos_policy(policy, ignore_missing=False)
                client.delete_qos_policy(obj)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete QoS policy "
                        "name or ID '%(qos_policy)s': %(e)s"
                    ),
                    {'qos_policy': policy, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.policy)
            msg = _(
                "%(result)s of %(total)s QoS policies failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListNetworkQosPolicy(command.Lister):
    _description = _("List QoS policies")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "List only QoS policies with the specified project "
                "(name or ID)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        shared_group = parser.add_mutually_exclusive_group()
        shared_group.add_argument(
            '--share',
            action='store_true',
            help=_("List only QoS policies shared between projects"),
        )
        shared_group.add_argument(
            '--no-share',
            action='store_true',
            help=_("List only QoS policies not shared between projects"),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        client = self.app.client_manager.network
        columns = (
            'id',
            'name',
            'is_shared',
            'is_default',
            'project_id',
        )
        column_headers = (
            'ID',
            'Name',
            'Shared',
            'Default',
            'Project',
        )

        attrs = _get_attrs(self.app.client_manager, parsed_args)

        if parsed_args.marker is not None:
            attrs['marker'] = parsed_args.marker
        if parsed_args.limit is not None:
            attrs['limit'] = parsed_args.limit
        if parsed_args.max_items is not None:
            attrs['max_items'] = parsed_args.max_items

        data = client.qos_policies(**attrs)
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class SetNetworkQosPolicy(common.NeutronCommandWithExtraArgs):
    _description = _("Set QoS policy properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar="<qos-policy>",
            help=_("QoS policy to modify (name or ID)"),
        )
        parser.add_argument(
            '--name', metavar="<name>", help=_('Set QoS policy name')
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_("Description of the QoS policy"),
        )
        enable_group = parser.add_mutually_exclusive_group()
        enable_group.add_argument(
            '--share',
            action='store_true',
            help=_('Make the QoS policy accessible by other projects'),
        )
        enable_group.add_argument(
            '--no-share',
            action='store_true',
            help=_('Make the QoS policy not accessible by other projects'),
        )
        default_group = parser.add_mutually_exclusive_group()
        default_group.add_argument(
            '--default',
            action='store_true',
            help=_("Set this as a default network QoS policy"),
        )
        default_group.add_argument(
            '--no-default',
            action='store_true',
            help=_("Set this as a non-default network QoS policy"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        client = self.app.client_manager.network
        obj = client.find_qos_policy(parsed_args.policy, ignore_missing=False)
        attrs = _get_attrs(self.app.client_manager, parsed_args)
        attrs.update(
            self._parse_extra_properties(parsed_args.extra_properties)
        )
        client.update_qos_policy(obj, **attrs)


class ShowNetworkQosPolicy(command.ShowOne):
    _description = _("Display QoS policy details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar="<qos-policy>",
            help=_("QoS policy to display (name or ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        client = self.app.client_manager.network
        obj = client.find_qos_policy(parsed_args.policy, ignore_missing=False)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters=_formatters)
        return (display_columns, data)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from openstack import exceptions as sdk_exc
from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import project as _project
from openstack.identity.v3 import user as _user
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.identity.v3 import project
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestProjectCreate(identity_fakes.TestIdentity):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    columns = (
        'description',
        'domain_id',
        'enabled',
        'id',
        'is_domain',
        'name',
        'options',
        'parent_id',
        'tags',
    )

    project_kwargs_no_options = {
        'description': None,
        'domain_id': None,
        'enabled': True,
        'is_domain': False,
        'parent_id': None,
        'tags': [],
    }

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_domain.return_value = self.domain

        # Get the command object to test
        self.cmd = project.CreateProject(self.app, None)

    def test_project_create_no_options(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project, **self.project_kwargs_no_options
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', project.name),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'name': project.name,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)

        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_description(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, description='new desc'),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--description',
            'new desc',
            project.name,
        ]
        verifylist = [
            ('description', 'new desc'),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'name': project.name,
            'description': 'new desc',
            'is_enabled': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            'new desc',
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_domain(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, domain_id=self.domain.id),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--domain',
            project.domain_id,
            project.name,
        ]
        verifylist = [
            ('domain', project.domain_id),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'domain_id': project.domain_id,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            self.domain.id,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_domain_no_perms(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, domain_id=self.domain.id),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--domain',
            project.domain_id,
            project.name,
        ]
        verifylist = [
            ('domain', project.domain_id),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.identity_sdk_client.find_domain.side_effect = (
            sdk_exc.ForbiddenException
        )
        self.identity_sdk_client.find_domain.return_value = None

        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'domain_id': project.domain_id,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            self.domain.id,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_enable(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, enabled=True),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--enable',
            project.name,
        ]
        verifylist = [
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_disable(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, enabled=False),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--disable',
            project.name,
        ]
        verifylist = [
            ('enabled', False),
            ('name', project.name),
            ('parent', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': False,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            False,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_property(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, fee='fi', fo='fum'),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--property',
            'fee=fi',
            '--property',
            'fo=fum',
            project.name,
        ]
        verifylist = [
            ('name', project.name),
            ('properties', {'fee': 'fi', 'fo': 'fum'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': True,
            'fee': 'fi',
            'fo': 'fum',
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(
            (
                'description',
                'domain_id',
                'enabled',
                'fee',
                'fo',
                'id',
                'is_domain',
                'name',
                'options',
                'parent_id',
                'tags',
            ),
            columns,
        )
        datalist = (
            None,
            None,
            True,
            'fi',
            'fum',
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_is_domain_false_property(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, is_domain=False),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--property',
            'is_domain=false',
            project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', project.name),
            ('tags', []),
            ('properties', {'is_domain': 'false'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': True,
            'is_domain': False,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_is_domain_true_property(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, is_domain=True),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--property',
            'is_domain=true',
            project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', project.name),
            ('tags', []),
            ('properties', {'is_domain': 'true'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': True,
            'is_domain': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            True,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_is_domain_none_property(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, is_domain=None),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--property',
            'is_domain=none',
            project.name,
        ]
        verifylist = [
            ('parent', None),
            ('enabled', True),
            ('name', project.name),
            ('tags', []),
            ('properties', {'is_domain': 'none'}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': True,
            'is_domain': None,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            None,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_parent(self):
        parent = sdk_fakes.generate_fake_resource(_project.Project)
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(
                self.project_kwargs_no_options,
                domain_id=self.domain.id,
                parent_id=parent.id,
            ),
        )
        self.identity_sdk_client.find_project.return_value = parent
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--domain',
            project.domain_id,
            '--parent',
            parent.name,
            project.name,
        ]
        verifylist = [
            ('domain', project.domain_id),
            ('parent', parent.name),
            ('enabled', True),
            ('name', project.name),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'name': project.name,
            'domain_id': project.domain_id,
            'parent_id': parent.id,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            self.domain.id,
            True,
            project.id,
            False,
            project.name,
            {},
            parent.id,
            [],
        )
        self.assertEqual(data, datalist)

    def test_project_create_invalid_parent(self):
        self.identity_sdk_client.find_project.side_effect = (
            sdk_exc.ResourceNotFound
        )
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(
                self.project_kwargs_no_options,
                domain_id=self.domain.id,
                parent_id='invalid',
            ),
        )

        arglist = [
            '--domain',
            project.domain_id,
            '--parent',
            'invalid',
            project.name,
        ]
        verifylist = [
            ('domain', project.domain_id),
            ('parent', 'invalid'),
            ('enabled', True),
            ('name', project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_project_create_with_tags(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(
                self.project_kwargs_no_options,
                domain_id=self.domain.id,
                tags=['foo'],
            ),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--domain',
            project.domain_id,
            '--tag',
            'foo',
            project.name,
        ]
        verifylist = [
            ('domain', project.domain_id),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', ['foo']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'domain_id': project.domain_id,
            'is_enabled': True,
            'tags': ['foo'],
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            self.domain.id,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            ['foo'],
        )
        self.assertEqual(datalist, data)

    def test_project_create_with_immutable_option(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(
                self.project_kwargs_no_options, options={'immutable': True}
            ),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--immutable',
            project.name,
        ]
        verifylist = [
            ('immutable', True),
            ('description', None),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': True,
            'options': {'immutable': True},
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {'immutable': True},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_with_no_immutable_option(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(
                self.project_kwargs_no_options, options={'immutable': False}
            ),
        )
        self.identity_sdk_client.create_project.return_value = project

        arglist = [
            '--no-immutable',
            project.name,
        ]
        verifylist = [
            ('immutable', False),
            ('description', None),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': project.name,
            'is_enabled': True,
            'options': {'immutable': False},
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {'immutable': False},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_conflict_with_or_show(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project, **self.project_kwargs_no_options
        )
        self.identity_sdk_client.create_project.side_effect = (
            sdk_exc.ConflictException
        )
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            '--or-show',
            project.name,
        ]
        verifylist = [
            ('or_show', True),
            ('description', None),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'name': project.name,
            'is_enabled': True,
        }
        self.identity_sdk_client.create_project.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_create_conflict_without_or_show(self):
        self.identity_sdk_client.create_project.side_effect = (
            sdk_exc.ConflictException
        )
        project = sdk_fakes.generate_fake_resource(
            _project.Project, **self.project_kwargs_no_options
        )

        arglist = [
            project.name,
        ]
        verifylist = [
            ('or_show', False),
            ('description', None),
            ('enabled', True),
            ('name', project.name),
            ('parent', None),
            ('tags', []),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            sdk_exc.ConflictException,
            self.cmd.take_action,
            parsed_args,
        )


class TestProjectDelete(identity_fakes.TestIdentity):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    def setUp(self):
        super().setUp()

        self.project = sdk_fakes.generate_fake_resource(_project.Project)
        self.project_with_domain = sdk_fakes.generate_fake_resource(
            _project.Project,
            name=self.project.name,
            domain_id=self.domain.id,
        )
        self.identity_sdk_client.delete_project.return_value = None

        # Get the command object to test
        self.cmd = project.DeleteProject(self.app, None)

    def test_project_delete_no_options(self):
        self.identity_sdk_client.find_project.return_value = self.project

        arglist = [
            self.project.id,
        ]
        verifylist = [
            ('projects', [self.project.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_project.assert_called_with(
            self.project.id,
        )
        self.assertIsNone(result)

    def test_project_multi_delete(self):
        self.identity_sdk_client.find_project.side_effect = [
            self.project,
            self.project_with_domain,
        ]
        arglist = [self.project.id, self.project_with_domain.id]
        verifylist = [
            ('projects', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_project.assert_has_calls(
            [
                mock.call(self.project.id),
                mock.call(self.project_with_domain.id),
            ]
        )
        self.assertIsNone(result)

    def test_project_delete_with_forbidden_domain(self):
        self.identity_sdk_client.find_domain.side_effect = [
            sdk_exc.ForbiddenException
        ]
        self.identity_sdk_client.find_project.return_value = (
            self.project_with_domain
        )

        arglist = [
            '--domain',
            self.project_with_domain.domain_id,
            self.project_with_domain.name,
        ]
        verifylist = [
            ('domain', self.domain.id),
            ('projects', [self.project_with_domain.name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.find_project.assert_called_with(
            name_or_id=self.project_with_domain.name,
            ignore_missing=False,
            domain_id=self.domain.id,
        )
        self.identity_sdk_client.delete_project.assert_called_once_with(
            self.project_with_domain.id
        )
        self.assertIsNone(result)

    def test_delete_multi_projects_with_exception(self):
        self.identity_sdk_client.find_project.side_effect = [
            self.project,
            self.project_with_domain,
            sdk_exc.NotFoundException,
        ]

        arglist = [
            self.project.id,
            self.project_with_domain.id,
            'unexist_project',
        ]
        verifylist = [
            ('projects', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 3 projects failed to delete.', str(e))

        self.identity_sdk_client.find_project.assert_has_calls(
            [
                mock.call(name_or_id=self.project.id, ignore_missing=False),
                mock.call(
                    name_or_id=self.project_with_domain.id,
                    ignore_missing=False,
                ),
                mock.call(name_or_id='unexist_project', ignore_missing=False),
            ]
        )

        self.assertEqual(3, self.identity_sdk_client.find_project.call_count)
        self.identity_sdk_client.delete_project.assert_has_calls(
            [
                mock.call(self.project.id),
                mock.call(self.project_with_domain.id),
            ]
        )


class TestProjectList(identity_fakes.TestIdentity):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)
    project = sdk_fakes.generate_fake_resource(
        _project.Project, domain_id=domain.id
    )
    projects = list(
        sdk_fakes.generate_fake_resources(_project.Project, count=2)
    )

    columns = (
        'ID',
        'Name',
    )
    datalist = (
        (
            project.id,
            project.name,
        ),
    )
    datalists = (
        (
            projects[0].description,
            True,
            projects[0].id,
            projects[0].name,
        ),
        (
            projects[1].description,
            True,
            projects[1].id,
            projects[1].name,
        ),
    )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = project.ListProject(self.app, None)

    def test_project_list_no_options(self):
        self.identity_sdk_client.projects.return_value = [self.project]

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.projects.assert_called_with()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_long(self):
        self.identity_sdk_client.projects.return_value = [self.project]

        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.projects.assert_called_with()

        collist = ('ID', 'Name', 'Domain ID', 'Description', 'Enabled')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.project.id,
                self.project.name,
                self.project.domain_id,
                self.project.description,
                True,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_project_list_domain(self):
        self.identity_sdk_client.projects.return_value = [self.project]

        arglist = [
            '--domain',
            self.project.domain_id,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
        ]

        self.identity_sdk_client.find_domain.return_value = self.domain

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.projects.assert_called_with(
            domain_id=self.project.domain_id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_domain_no_perms(self):
        self.identity_sdk_client.projects.return_value = [self.project]

        arglist = [
            '--domain',
            self.project.domain_id,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.identity_sdk_client.find_project.side_effect = (
            sdk_exc.ResourceNotFound
        )
        self.identity_sdk_client.find_domain.return_value = self.domain

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.projects.assert_called_with(
            domain_id=self.project.domain_id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_parent(self):
        self.parent = sdk_fakes.generate_fake_resource(_project.Project)
        self.project = sdk_fakes.generate_fake_resource(
            _project.Project,
            id=self.project.id,
            name=self.project.name,
            domain_id=self.domain.id,
            parent_id=self.parent.id,
        )
        self.identity_sdk_client.projects.return_value = [self.project]

        arglist = [
            '--parent',
            self.parent.id,
        ]
        verifylist = [
            ('parent', self.parent.id),
        ]

        self.identity_sdk_client.find_project.return_value = self.parent

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.projects.assert_called_with(
            parent_id=self.parent.id
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_user(self):
        self.user = sdk_fakes.generate_fake_resource(_user.User)
        self.project = sdk_fakes.generate_fake_resource(
            _project.UserProject,
            id=self.project.id,
            name=self.project.name,
            user_id=self.user.id,
        )
        self.identity_sdk_client.user_projects.return_value = [self.project]

        arglist = [
            '--user',
            self.user.id,
        ]
        verifylist = [
            ('user', self.user.id),
        ]

        self.identity_sdk_client.find_user.return_value = self.user

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.user_projects.assert_called_with(self.user.id)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_sort(self):
        self.identity_sdk_client.projects.return_value = self.projects

        arglist = [
            '--sort',
            'name:asc',
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        (columns, data) = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.projects.assert_called_with()

        collist = ('ID', 'Name')
        self.assertEqual(collist, columns)

        if self.projects[0].name > self.projects[1].name:
            datalists = (
                (self.projects[1].id, self.projects[1].name),
                (self.projects[0].id, self.projects[0].name),
            )
        else:
            datalists = (
                (self.projects[0].id, self.projects[0].name),
                (self.projects[1].id, self.projects[1].name),
            )

        self.assertEqual(datalists, tuple(data))

    def test_project_list_my_projects(self):
        self.identity_sdk_client.user_projects.return_value = [self.project]

        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN_WITH_PROJECT_ID,
        )
        self.app.client_manager.auth_ref = auth_ref

        arglist = [
            '--my-projects',
        ]
        verifylist = [
            ('my_projects', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.user_projects.assert_called_with(
            self.app.client_manager.auth_ref.user_id
        )

        collist = ('ID', 'Name')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.project.id,
                self.project.name,
            ),
        )
        self.assertEqual(datalist, tuple(data))

    def test_project_list_with_option_enabled(self):
        self.identity_sdk_client.projects.return_value = [self.project]

        arglist = ['--enabled']
        verifylist = [('is_enabled', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'is_enabled': True}
        self.identity_sdk_client.projects.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_project_list_with_pagination(self):
        self.identity_sdk_client.projects.return_value = [self.project]

        arglist = [
            '--limit',
            '2',
            '--marker',
            'some-marker',
        ]
        verifylist = [
            ('limit', 2),
            ('marker', 'some-marker'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'limit': 2, 'marker': 'some-marker'}
        self.identity_sdk_client.projects.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))


class TestProjectSet(identity_fakes.TestIdentity):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    project_kwargs_no_options = {
        'domain_id': domain.id,
        'tags': ['tag1', 'tag2', 'tag3'],
    }
    project = sdk_fakes.generate_fake_resource(
        _project.Project, **project_kwargs_no_options
    )

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.find_project.return_value = self.project

        # Get the command object to test
        self.cmd = project.SetProject(self.app, None)

    def test_project_set_no_options(self):
        arglist = [
            self.project.name,
        ]
        verifylist = [
            ('project', self.project.name),
            ('enabled', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)

    def test_project_set_name(self):
        arglist = [
            '--name',
            'qwerty',
            '--domain',
            self.project.domain_id,
            self.project.name,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('domain', self.project.domain_id),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'qwerty',
        }

        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_description(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--description',
            'new desc',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('description', 'new desc'),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': 'new desc',
        }
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_enable(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--enable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('enabled', True),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': True,
        }
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_disable(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--disable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('enabled', False),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'enabled': False,
        }
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_property(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--property',
            'fee=fi',
            '--property',
            'fo=fum',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('properties', {'fee': 'fi', 'fo': 'fum'}),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'fee': 'fi',
            'fo': 'fum',
        }
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_tags(self):
        arglist = [
            '--name',
            'qwerty',
            '--domain',
            self.project.domain_id,
            '--tag',
            'foo',
            self.project.name,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('domain', self.domain.id),
            ('enabled', None),
            ('project', self.project.name),
            ('tags', ['foo']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values. new tag is added to original tags for update.
        kwargs = {
            'name': 'qwerty',
            'tags': sorted({'tag1', 'tag2', 'tag3', 'foo'}),
        }
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_remove_tags(self):
        arglist = [
            '--remove-tag',
            'tag1',
            '--remove-tag',
            'tag2',
            self.project.name,
        ]
        verifylist = [
            ('enabled', None),
            ('project', self.project.name),
            ('remove_tags', ['tag1', 'tag2']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {'tags': list({'tag3'})}
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_with_immutable_option(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--immutable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('immutable', True),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': True},
        }
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)

    def test_project_set_with_no_immutable_option(self):
        arglist = [
            '--domain',
            self.project.domain_id,
            '--no-immutable',
            self.project.name,
        ]
        verifylist = [
            ('domain', self.project.domain_id),
            ('immutable', False),
            ('enabled', None),
            ('project', self.project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': False},
        }
        self.identity_sdk_client.update_project.assert_called_with(
            self.project.id, **kwargs
        )
        self.assertIsNone(result)


class TestProjectShow(identity_fakes.TestIdentity):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    columns = (
        'description',
        'domain_id',
        'enabled',
        'id',
        'is_domain',
        'name',
        'options',
        'parent_id',
        'tags',
    )

    project_kwargs_no_options = {
        'description': None,
        'domain_id': None,
        'enabled': True,
        'is_domain': False,
        'parent_id': None,
        'tags': [],
    }

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = project.ShowProject(self.app, None)

    def test_project_show(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project, **self.project_kwargs_no_options
        )
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_project.assert_called_with(
            project.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_show_parents(self):
        parent = sdk_fakes.generate_fake_resource(
            _project.Project, parent_id='default'
        )
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(
                self.project_kwargs_no_options,
                parent_id=parent.id,
                parents={parent.id: {parent.parent_id: None}},
            ),
        )
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            project.id,
            '--parents',
        ]
        verifylist = [
            ('project', project.id),
            ('parents', True),
            ('children', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_project.assert_called_with(
            project.id, parents_as_ids=True, ignore_missing=False
        )

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'options',
            'parent_id',
            'parents',
            'tags',
        )
        self.assertEqual(collist, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            parent.id,
            {parent.id: {'default': None}},
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_show_subtree(self):
        child = sdk_fakes.generate_fake_resource(
            _project.Project, subtree=None
        )
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, subtree={child.id: None}),
        )
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            project.id,
            '--children',
        ]
        verifylist = [
            ('project', project.id),
            ('parents', False),
            ('children', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_project.assert_called_with(
            project.id, subtree_as_ids=True, ignore_missing=False
        )

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'options',
            'parent_id',
            'subtree',
            'tags',
        )
        self.assertEqual(collist, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            {child.id: None},
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_show_parents_and_children(self):
        parent = sdk_fakes.generate_fake_resource(
            _project.Project, parent_id='default'
        )
        child = sdk_fakes.generate_fake_resource(
            _project.Project, subtree=None
        )
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(
                self.project_kwargs_no_options,
                parent_id=parent.id,
                parents={parent.id: {parent.parent_id: None}},
                subtree={child.id: None},
            ),
        )
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            project.id,
            '--parents',
            '--children',
        ]
        verifylist = [
            ('project', project.id),
            ('parents', True),
            ('children', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_project.assert_called_with(
            project.id,
            parents_as_ids=True,
            subtree_as_ids=True,
            ignore_missing=False,
        )

        collist = (
            'description',
            'domain_id',
            'enabled',
            'id',
            'is_domain',
            'name',
            'options',
            'parent_id',
            'parents',
            'subtree',
            'tags',
        )
        self.assertEqual(collist, columns)
        datalist = (
            None,
            None,
            True,
            project.id,
            False,
            project.name,
            {},
            parent.id,
            {parent.id: {'default': None}},
            {child.id: None},
            [],
        )
        self.assertEqual(datalist, data)

    def test_project_show_with_domain(self):
        project = sdk_fakes.generate_fake_resource(
            _project.Project,
            **dict(self.project_kwargs_no_options, domain_id=self.domain.id),
        )
        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.find_project.return_value = project

        arglist = [
            "--domain",
            self.domain.id,
            project.name,
        ]
        verifylist = [
            ('domain', self.domain.id),
            ('project', project.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_project.assert_called_with(
            project.id, domain_id=self.domain.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            None,
            self.domain.id,
            True,
            project.id,
            False,
            project.name,
            {},
            None,
            [],
        )
        self.assertEqual(datalist, data)



# Copyright (c) 2018 Orange SA.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import operator
from unittest import mock

from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from openstackclient.network.v2.bgpvpn import router_association
from openstackclient.tests.unit.network.v2.bgpvpn import fakes
from openstackclient.tests.unit import utils


columns_short = tuple(
    col
    for col, _, listing_mode in router_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_SHORT_ONLY)
)
columns_long = tuple(
    col
    for col, _, listing_mode in router_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
headers_short = tuple(
    head
    for _, head, listing_mode in router_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_SHORT_ONLY)
)
headers_long = tuple(
    head
    for _, head, listing_mode in router_association._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
sorted_attr_map = sorted(
    router_association._attr_map, key=operator.itemgetter(1)
)
sorted_columns = tuple(col for col, _, _ in sorted_attr_map)
sorted_headers = tuple(head for _, head, _ in sorted_attr_map)


def _get_data(attrs, columns=sorted_columns):
    return osc_utils.get_dict_properties(
        attrs, columns, formatters=router_association._formatters
    )


class TestCreateRouterAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = router_association.CreateBgpvpnRouterAssoc(self.app, None)
        self.fake_bgpvpn = fakes.create_one_bgpvpn()
        self.fake_router = fakes.create_one_resource()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.network_client.find_router.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.fake_project = mock.Mock(id='fake_project_id')
        self.identity_sdk_client.find_project = mock.Mock(
            return_value=self.fake_project
        )

    def _build_args(self, param=None):
        arglist_base = [
            self.fake_bgpvpn['id'],
            self.fake_router['id'],
            '--project',
            self.fake_bgpvpn['project_id'],
        ]
        if param is not None:
            if isinstance(param, list):
                arglist_base.extend(param)
            else:
                arglist_base.append(param)
        return arglist_base

    def _build_verify_list(self, param=None):
        verifylist = [
            ('bgpvpn', self.fake_bgpvpn['id']),
            ('resource', self.fake_router['id']),
            ('project', self.fake_bgpvpn['project_id']),
        ]
        if param is not None:
            verifylist.append(param)
        return verifylist

    def _exec_create_router_association(
        self,
        fake_assoc,
        arglist,
        verifylist,
    ):
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        cols, data = self.cmd.take_action(parsed_args)

        fake_assoc_call = {
            'router_id': self.fake_router['id'],
            'project_id': 'fake_project_id',
        }
        if verifylist:
            for key, value in verifylist:
                if (
                    key
                    not in (
                        'bgpvpn',
                        'resource',
                        'project',
                    )
                    and value not in fake_assoc_call.values()
                ):
                    fake_assoc_call[key] = value
        self.network_client.create_bgpvpn_router_association.assert_called_once_with(
            self.fake_bgpvpn['id'], **fake_assoc_call
        )
        return cols, data

    def test_create_router_association(self):
        fake_assoc = fakes.create_one_router_association()

        self.network_client.create_bgpvpn_router_association.return_value = (
            fake_assoc
        )

        arglist = self._build_args()
        # advertise_extra_routes will be False since none
        # of the mutually exclusive args present
        verifylist = self._build_verify_list(('advertise_extra_routes', False))

        self._exec_create_router_association(fake_assoc, arglist, verifylist)

    def test_create_router_association_advertise(self):
        fake_assoc = fakes.create_one_router_association(
            {'advertise_extra_routes': True}
        )

        self.network_client.create_bgpvpn_router_association.return_value = (
            fake_assoc
        )

        arglist = self._build_args('--advertise_extra_routes')
        verifylist = self._build_verify_list(('advertise_extra_routes', True))

        cols, data = self._exec_create_router_association(
            fake_assoc, arglist, verifylist
        )
        self.assertEqual(sorted(sorted_columns), sorted(cols))
        self.assertEqual(_get_data(fake_assoc, cols), data)

    def test_create_router_association_no_advertise(self):
        fake_assoc = fakes.create_one_router_association(
            {'advertise_extra_routes': False}
        )

        self.network_client.create_bgpvpn_router_association.return_value = (
            fake_assoc
        )

        arglist = self._build_args('--no-advertise_extra_routes')
        verifylist = self._build_verify_list(('advertise_extra_routes', False))

        cols, data = self._exec_create_router_association(
            fake_assoc, arglist, verifylist
        )
        self.assertEqual(sorted(sorted_columns), sorted(cols))
        self.assertEqual(_get_data(fake_assoc, cols), data)

    def test_create_router_association_advertise_fault(self):
        arglist = self._build_args(
            ['--advertise_extra_routes', '--no-advertise_extra_routes']
        )

        try:
            self._exec_create_router_association(None, arglist, None)
        except utils.ParserException as e:
            self.assertIn('Argument parse failed', format(e))

    def test_router_association_unknown_arg(self):
        arglist = self._build_args('--unknown arg')

        try:
            self._exec_create_router_association(None, arglist, None)
        except utils.ParserException as e:
            self.assertIn('Argument parse failed', format(e))


class TestSetRouterAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = router_association.SetBgpvpnRouterAssoc(self.app, None)
        self.fake_bgpvpn = fakes.create_one_bgpvpn()
        self.fake_router = fakes.create_one_resource()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )

    def _build_args(self, fake_assoc, param=None):
        arglist_base = [fake_assoc['id'], self.fake_bgpvpn['id']]
        if param is not None:
            if isinstance(param, list):
                arglist_base.extend(param)
            else:
                arglist_base.append(param)
        return arglist_base

    def test_set_router_association_no_advertise(self):
        fake_assoc = fakes.create_one_router_association(
            {'advertise_extra_routes': True}
        )

        arglist = self._build_args(fake_assoc, '--no-advertise_extra_routes')
        verifylist = [
            ('resource_association_id', fake_assoc['id']),
            ('bgpvpn', self.fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_bgpvpn_router_association.assert_called_once_with(
            self.fake_bgpvpn['id'],
            fake_assoc['id'],
            **{'advertise_extra_routes': False},
        )
        self.assertIsNone(result)

    def test_set_router_association_advertise(self):
        fake_assoc = fakes.create_one_router_association(
            {'advertise_extra_routes': False}
        )

        arglist = self._build_args(fake_assoc, '--advertise_extra_routes')
        verifylist = [
            ('resource_association_id', fake_assoc['id']),
            ('bgpvpn', self.fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network_client.update_bgpvpn_router_association.assert_called_once_with(
            self.fake_bgpvpn['id'],
            fake_assoc['id'],
            **{'advertise_extra_routes': True},
        )
        self.assertIsNone(result)


class TestShowRouterAssoc(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = router_association.ShowBgpvpnRouterAssoc(self.app, None)
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )

    def test_show_router_association(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        fake_assoc = fakes.create_one_router_association(
            {'advertise_extra_routes': True}
        )
        self.network_client.get_bgpvpn_router_association.return_value = (
            fake_assoc
        )
        arglist = [
            fake_assoc['id'],
            fake_bgpvpn['id'],
        ]
        verifylist = [
            ('resource_association_id', fake_assoc['id']),
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        cols, data = self.cmd.take_action(parsed_args)

        self.network_client.get_bgpvpn_router_association.assert_called_once_with(
            fake_bgpvpn['id'], fake_assoc['id']
        )
        self.assertEqual(sorted(sorted_columns), sorted(cols))
        self.assertEqual(data, _get_data(fake_assoc, cols))
